import re
from typing import List, Tuple, Optional, cast
import sys

import numpy as np
from PIL import Image, UnidentifiedImageError
from rclip import utils

QUERY_WITH_MULTIPLIER_RE = re.compile(r'^(?P<multiplier>(\d+(\.\d+)?|\.\d+|\d+\.)):(?P<query>.+)$')
QueryWithMultiplier = Tuple[float, str]


class Model:
  VECTOR_SIZE = 512
  _model_name = 'ViT-B-32'
  _checkpoint_name = 'openai'

  def __init__(self, device: str = 'cpu'):
    self._device = device
    self.__model = None
    self.__preprocess = None
    self.__tokenizer = None

  @property
  def _tokenizer(self):
    import open_clip
    if not self.__tokenizer:
      self.__tokenizer = open_clip.get_tokenizer(self._model_name)
    return self.__tokenizer

  @property
  def _model(self):
    import open_clip
    if not self.__model:
      self.__model, _, self.__preprocess = open_clip.create_model_and_transforms(
        self._model_name,
        pretrained=self._checkpoint_name,
        device=self._device,
      )
    return self.__model

  @property
  def _preprocess(self):
    import open_clip
    if not self.__preprocess:
      self.__model, _, self.__preprocess = open_clip.create_model_and_transforms(
        self._model_name,
        pretrained=self._checkpoint_name,
        device=self._device,
      )
    return self.__preprocess

  def compute_image_features(self, images: List[Image.Image]) -> np.ndarray:
    import torch
    images_preprocessed = torch.stack([self._preprocess(thumb) for thumb in images]).to(self._device)
    with torch.no_grad():
      image_features = self._model.encode_image(images_preprocessed)
      image_features /= image_features.norm(dim=-1, keepdim=True)
    return image_features.cpu().numpy()

  def compute_text_features(self, text: List[str]) -> np.ndarray:
    import torch
    with torch.no_grad():
      text_features = self._model.encode_text(self._tokenizer(text).to(self._device))
      text_features /= text_features.norm(dim=-1, keepdim=True)
    return text_features.cpu().numpy()

  @staticmethod
  def _extract_query_multiplier(query: str) -> QueryWithMultiplier:
    match = QUERY_WITH_MULTIPLIER_RE.match(query)
    if not match:
      return 1., query
    multiplier = float(match.group('multiplier'))
    query = match.group('query')
    return multiplier, query

  @staticmethod
  def _group_queries_by_type(queries: List[str]) -> Tuple[
    List[QueryWithMultiplier],
    List[QueryWithMultiplier],
    List[QueryWithMultiplier]
  ]:
    phrase_queries: List[Tuple[float, str]] = []
    local_file_queries: List[Tuple[float, str]] = []
    url_queries: List[Tuple[float, str]] = []
    for query in queries:
        multiplier, query = Model._extract_query_multiplier(query)
        if utils.is_http_url(query):
          url_queries.append((multiplier, query))
        elif utils.is_file_path(query):
          local_file_queries.append((multiplier, query))
        else:
          phrase_queries.append((multiplier, query))
    return phrase_queries, local_file_queries, url_queries

  def compute_features_for_queries(self, queries: List[str]) -> np.ndarray:
    text_features: Optional[np.ndarray] = None
    image_features: Optional[np.ndarray] = None
    phrases, files, urls = self._group_queries_by_type(queries)
    if phrases:
      phrase_multipliers, phrase_queries = cast(Tuple[Tuple[float], Tuple[str]], zip(*phrases))
      phrase_multipliers_np = np.array(phrase_multipliers).reshape(-1, 1)
      text_features = np.add.reduce(self.compute_text_features([*phrase_queries]) * phrase_multipliers_np)
    if files or urls:
      file_multipliers, file_paths = cast(Tuple[Tuple[float], Tuple[str]], zip(*(files))) if files else ((), ())
      url_multipliers, url_paths = cast(Tuple[Tuple[float], Tuple[str]], zip(*(urls))) if urls else ((), ())
      try:
        images = ([utils.download_image(q) for q in url_paths] +
                  [utils.read_image(q) for q in file_paths])
      except FileNotFoundError as e:
        print(f'File "{e.filename}" not found. Check if you have typos in the filename.')
        sys.exit(1)
      except UnidentifiedImageError as e:
        print(f'File "{e.filename}" is not an image. You can only use image files or text as queries.')
        sys.exit(1)
      image_multipliers = np.array(url_multipliers + file_multipliers)
      image_features = np.add.reduce(self.compute_image_features(images) * image_multipliers.reshape(-1, 1))

    if text_features is not None and image_features is not None:
        return text_features + image_features
    elif text_features is not None:
        return text_features
    elif image_features is not None:
        return image_features
    else:
        return np.zeros(Model.VECTOR_SIZE)

  def compute_similarities_to_text(
      self, item_features: np.ndarray,
      positive_queries: List[str], negative_queries: List[str]) -> List[Tuple[float, int]]:

    positive_features = self.compute_features_for_queries(positive_queries)
    negative_features = self.compute_features_for_queries(negative_queries)

    features = positive_features - negative_features

    similarities = features @ item_features.T
    sorted_similarities = sorted(zip(similarities, range(item_features.shape[0])), key=lambda x: x[0], reverse=True)

    return sorted_similarities



import pathlib
import sqlite3
from typing import Any, Optional, TypedDict, Union


class ImageOmittable(TypedDict, total=False):
  deleted: bool


class NewImage(ImageOmittable):
  filepath: str
  modified_at: float
  size: int
  vector: bytes


class Image(NewImage):
  id: int


class DB:
  VERSION = 1

  def __init__(self, filename: Union[str, pathlib.Path]):
    self._con = sqlite3.connect(filename)
    self._con.row_factory = sqlite3.Row
    self.ensure_tables()
    self.ensure_version()

  def ensure_tables(self):
    self._con.execute('''
      CREATE TABLE IF NOT EXISTS images (
        id INTEGER PRIMARY KEY,
        deleted BOOLEAN,
        filepath TEXT NOT NULL UNIQUE,
        modified_at DATETIME NOT NULL,
        size INTEGER NOT NULL,
        vector BLOB NOT NULL
      )
    ''')
    # Query for images
    self._con.execute('CREATE UNIQUE INDEX IF NOT EXISTS existing_images ON images(filepath) WHERE deleted IS NULL')

    self._con.execute('CREATE TABLE IF NOT EXISTS db_version (version INTEGER)')

    self._con.commit()

  def ensure_version(self):
    db_version = self._con.execute('SELECT version FROM db_version').fetchone()
    if not db_version:
      self._con.execute('INSERT INTO db_version(version) VALUES (?)', (self.VERSION,))
      self._con.commit()
    elif db_version['version'] < self.VERSION:
      raise Exception('migration to a newer index version isn\'t implemented')
    elif db_version['version'] > self.VERSION:
      raise Exception(
        'found index version newer than this version of rclip can support;'
        ' please, update rclip: https://github.com/yurijmikhalevich/rclip/releases'
      )

  def commit(self):
    self._con.commit()

  def upsert_image(self, image: NewImage, commit: bool = True):
    self._con.execute('''
      INSERT INTO images(deleted, filepath, modified_at, size, vector)
      VALUES (:deleted, :filepath, :modified_at, :size, :vector)
      ON CONFLICT(filepath) DO UPDATE SET
        deleted=:deleted, modified_at=:modified_at, size=:size, vector=:vector
    ''', {'deleted': None, **image})
    if commit:
      self._con.commit()

  def flag_images_in_a_dir_as_deleted(self, path: str):
    self._con.execute('UPDATE images SET deleted = 1 WHERE filepath LIKE ?', (path + '/%',))
    self._con.commit()

  def remove_deleted_flag(self, filepath: str, commit: bool = True):
    self._con.execute('UPDATE images SET deleted = NULL WHERE filepath = ?', (filepath,))
    if commit:
      self._con.commit()

  def get_image(self, **kwargs: Any) -> Optional[Image]:
    query = ' AND '.join(f'{key}=:{key}' for key in kwargs)
    cur = self._con.execute(f'SELECT * FROM images WHERE {query} LIMIT 1', kwargs)
    return cur.fetchone()

  def get_image_vectors_by_dir_path(self, path: str) -> sqlite3.Cursor:
    return self._con.execute(
      f'SELECT filepath, vector FROM images WHERE filepath LIKE ? AND deleted IS NULL', (path + '/%',)
    )
